/***************************************************************************/
/*                                                                         */
/* FILE NAME : ga-for-gp.h                                                 */
/*                                                                         */
/*                                                                         */
/***************************************************************************/

/*
  Program for graph partitioning using GA's, written by Dr. Ching-Tai Chiu 
*/

#include "ga.h"
#include <stdio.h>

extern void generate_initial_set();
extern void generate_graph();
extern double calc_pop_fitness(), calc_fitness();
extern void reproduction();
extern void rescale();
int evol, spop, besti;
double zs[MAX_POP];       /* Fitness value of this solution */ 
double w[MAX_ND][MAX_ND]; /* Weights between graph nodes */
double wS[MAX_ND][MAX_ND]; /* Starting Weights between graph nodes */
float p[MAX_POP];
int nodecount;
int bstr[MAX_POP][MAX_ND], bestsol[MAX_ND];
FILE *fp;

DebugPrint(str)
char *str;
{ /*printf("\n %s \n",str);
  */
  return;
}

/*
* Finds the minimum valued element's index from the double array.
*/
int Find_minimum(darray,size)
double darray[];
int size;
{  int i;
   int min = 0;
   for(i=1;i<size;i++)
      if(darray[min]>darray[i]) min=i;
   return min;
}/* Find_minimum() */

/*
* Prints the poupulation information. Each solution. Each solution's
* fitness value and if it has the minimum fitness value then this solution
* has a double star.
*/
PrintPopulation(fp, gen_num, nodecount, pop_size)
FILE *fp;
int gen_num;
int nodecount;
int pop_size;
{  int i,j; 
   int min;

   min=Find_minimum(zs,pop_size);

   fprintf(fp,"\n Generation : %d ",gen_num);

   for(i=0;i<pop_size;i++)
   {  fprintf(fp,"\n Solution %3d -->",i); 
      for(j=0;j<nodecount;j++) fprintf(fp, "%d", bstr[i][j]);
      fprintf(fp,"  %8.4f ",zs[i]);
      if((i==min)||(zs[i]==zs[min])) fprintf(fp," ** ");
   }
   fprintf(fp,"\n");
}/* end of PrintPopulation() */

main(argc, argv)
int argc;
char *argv[];
{
  int i, j, worsti, count0, count1;
  double avr_fit, zmax;
  long int clock;
  FILE *mpx, *mpy;
  char mfx[30], mfy[30];

  if (argc==1)
    {
      printf("Usage:\n");
      printf("num_nodes population epochs outputfile\n");
      exit(0);
    }
  sscanf(argv[1],"%d", &nodecount);
/*  nodecount = atoi(argv[1]);  */
  sscanf(argv[2], "%d", &spop);
/*  spop = atoi(argv[2]);  */
  sscanf(argv[3], "%d", &evol);
/*  evol = atoi(argv[3]);  */

  if ((fp = fopen(argv[4], "w+t")) == NULL)
    {
      printf("Could not open output file\n");
      exit(0);
    }

  DebugPrint("I opened the output file ");

  rewind(fp);

  strcpy(mfx, argv[4]);
  strcpy(mfy, argv[4]);
  strcat(mfx, "x.mat");
  strcat(mfy, "y.mat");
  mpx = fopen(mfx, "w+t");
  mpy = fopen(mfy, "w+t");

  /*generate_graph(); */
  generate_graph_write("Graph");

  DebugPrint(" I generated GRAPH "); 

  clock = time(0);
  generate_initial_set();
  avr_fit = calc_pop_fitness();
  rescale();

  DebugPrint(" Let's print out what we got so far "); 

  for (i=0; i<spop; i++) {
    fprintf(fp, "\n%d    %6.2f  %6.3f --- ", i, zs[i], p[i]);
    for (j=0; j<nodecount; j++) fprintf(fp, "%d", bstr[i][j]);
  }

  fprintf(fp, "\nAverage fitness = %8.4f    Min[%d]=%6.3f\n", 
	  avr_fit, besti, zs[besti]);

  fprintf(mpx, "%d\n", 0);
  fprintf(mpy, "%5.0f\n", zs[besti]);

  /* HERE main generation starts. */

  for (i=0; i<evol; i++) 
  {
     PrintPopulation(fp, i, nodecount, spop);
     DebugPrint("EVOL ");
     printf("\n evol : %d",i);

     reproduction();

     DebugPrint("EVOL1");

     /*   zmax = zs[0];
          worsti = 0;
          for (j=0; j<spop; j++) 
            if (zmax < zs[j]) 
	    {   worsti = j;
	        zmax = zs[j];
            }
          for (j=0; j<nodecount; j++)	bstr[worsti][j] = bestsol[j];
     */

     avr_fit = calc_pop_fitness();

     printf("Average fitness at %d is %8.3f ",i,avr_fit);

     DebugPrint("EVOL2");

     rescale();

     DebugPrint("EVOL3");

     printf("\nGeneration %4d: average fitness = %8.4f    Min[%d]=%6.3f\n", 
	   i+1, avr_fit, besti, zs[besti]);

     fprintf(mpx, "%d\n", i+1);
     fprintf(mpy, "%5.0f\n", zs[besti]);

     if ((i % 10) == 0)
       fprintf(fp, "\n%5d: Average fitness = %8.4f  Min[%d]=%6.3f", 
	      i+1, avr_fit, besti, zs[besti]);   


     if (drand48()<0.02 || (avr_fit - zs[besti]) < 0.005) 
     {
        printf("\nMutation!!");
        mutation();
        avr_fit = calc_pop_fitness();
        rescale();
        i++;
        fprintf(fp, "\n%5d: Average fitness = %8.4f  Min[%d]=%6.3f", 
	      i+1, avr_fit, besti, zs[besti]);
        fprintf(mpx, "%d\n", i+1);
        fprintf(mpy, "%5.0f\n", zs[besti]);

        /*   zmax = zs[0];
             worsti = 0;
             for (j=0; j<spop; j++) 
	     if (zmax < zs[j]) 
	     {
	        worsti = j;
                zmax = zs[j];
	     }
             for (j=0; j<nodecount; j++) bstr[worsti][j] = bestsol[j];  
	*/

    }/* end of if */

  }/* end of for */

  clock = time(0) - clock;
  for (i=0; i<spop; i++) {
    fprintf(fp, "\n%d    %6.2f  %6.3f --- ", i, zs[i], p[i]);
    for (j=0; j<nodecount; j++) fprintf(fp, "%d", bstr[i][j]);
  }

  fprintf(fp, "\nNodes = %d\nPopulation = %d\nGenerations = %d",
	  nodecount, spop, evol);

  fprintf(fp, "\n\nBest cost = %8.2f -> ", calc_fitness(bstr[besti]));
  count0 = count1 = 0;
  for (i=0; i<nodecount; i++) 
    {
      fprintf(fp, "%d", bstr[besti][i]);
      if (bstr[besti][i] == 0)
	count0++;
      else
	count1++;
    }
  fprintf(fp, "\n\n%d nodes in 0 set\n%d nodes in 1 set\n", count0, count1);

  fprintf(fp, "\nTime = %d\n", clock);

  fclose(mpx);
  fclose(mpy);
  fclose(fp);
}
